/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Authors: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>
             Michael Tross, digitalSTROM AG <michael.tross@digitalstrom.com>
             Reto Flütsch, digitalSTROM AG <reto.fluetsch@digitalstrom.com>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#include "group.h"
#include "scene.h"
#include "dsm.h"
#include "selectA.h"
#include "datamodel.h"
#include "registers.h"
#include "bsp.h"
#include "button.h"

Q_DEFINE_THIS_MODULE(group)

/* FSM state-handler functions */
static QState Group_initial(Group *me, QEvent const *e);

static QState Group_room_sensitive(Group *me, QEvent const *e);
static QState Group_room_on(Group *me, QEvent const *e);
static QState Group_room_dimming(Group *me, QEvent const *e);
static QState Group_room_on_dim_down(Group *me, QEvent const *e);
static QState Group_room_on_dimming_down(Group *me, QEvent const *e);
static QState Group_room_on_dim_up(Group *me, QEvent const *e);
static QState Group_room_on_dimming_up(Group *me, QEvent const *e);
static QState Group_off(Group *me, QEvent const *e);

static QState Group_on(Group *me, QEvent const *e);
static QState Group_on_register_local_prio(Group *me, QEvent const *e);
static QState Group_on_local(Group *me, QEvent const *e);
static QState Group_on_local_prio(Group *me, QEvent const *e);

static QState Group_local_prio(Group *me, QEvent const *e);
static QState Group_dim_down(Group *me, QEvent const *e);
static QState Group_dimming_down(Group *me, QEvent const *e);
static QState Group_dim_up(Group *me, QEvent const *e);
static QState Group_dimming_up(Group *me, QEvent const *e);

#define GetStateHandlerOffsetFromState(state_)  ((QStateHandler)state_)

static QStateHandler GroupOffStates[2];

bool Group_is_off_state(const QStateHandler state) {
    uint8_t i;
    for (i = 0; i < (sizeof(GroupOffStates) / sizeof(uint16_t)); ++i) {
        if (state == GroupOffStates[i]) {
            return TRUE;
        }
    }
    return FALSE;
}

static void Group_RestartSTOPTimer(Group *me);
static void Group_StopSTOPTimer(Group *me);
static bool Group_SendSTOP(Group *me, QEvent const *e);

/** Duration of a long downstream message */
#define SHORT_DOWNSTREAM_MESSAGE_MS         (240)

/* timeout for generating internally a INC/DEC signal */
#define LIGHT_DIM_TIMEOUT_TICKS             (250/BSP_MS_PER_TICKS)

/* timeout for first INC/DEC signal after calling min dim */
#define LIGHT_SHORT_MESSAGE_TIMEOUT_TICKS   (SHORT_DOWNSTREAM_MESSAGE_MS/BSP_MS_PER_TICKS)

/* timeout for ending hold */
#define LIGHT_HOLD_END_TIMEOUT_TICKS        (2100/BSP_MS_PER_TICKS)

/* timeout for setting local priority */
#define LIGHT_SET_LOCAL_PRIO_TIMEOUT_TICKS  (2000/BSP_MS_PER_TICKS)

static void Group_RestartSTOPTimer(Group *me)
{
    uint8_t stopTimeout = Datamodel_GetStopTimeoutZoneGroup(me->localZoneId, me->groupId);
    if (stopTimeout != 0) {
        if (me->stopTimer != NULL) {
            QTimeEvt_disarm(&me->stopTimer->super);
        } else {
            me->stopTimer = SelectA_AllocateStopTimer(me->localZoneId, me->groupId, me->buttonId);
            if (me->stopTimer == NULL) {
                /* TODO: send error event if no more resources are available */
                return;
            }
        }
        QTimeEvt_postIn(&me->stopTimer->super, me->parent, stopTimeout * 1000 / BSP_MS_PER_TICKS);
    }
}

static void Group_StopSTOPTimer(Group *me)
{
    if (me->stopTimer != NULL) {
        SelectA_ReleaseStopTimer(me->stopTimer);
        me->stopTimer = NULL;
    }
}

static bool Group_SendSTOP(Group *me, QEvent const *e)
{
    if (SelectA_RoomTimerRunning(me->localZoneId, me->groupId)) {
        /* re-post event for main HSM */
        ClickForwardEvt * newEvent = Q_NEW(ClickForwardEvt, INTERNAL_CLICK_FORWARD_SIG);
        newEvent->groupId = me->groupId;
        newEvent->localZoneId = me->localZoneId;
        newEvent->event = e->sig;
        newEvent->buttonId = (me->buttonId >= BUTTON_GROUP_5) ? (me->buttonId - 4) : 0;
        newEvent->fid = me->fid;
        newEvent->key = me->key;
        newEvent->deviceId = me->deviceId;
        newEvent->apartmentWide = FALSE;
        newEvent->outputChannel = me->outputChannel;
        QActive_postLIFO(me->parent, (QEvent *)newEvent);
        return TRUE;
    } else if (me->stopTimer != NULL) {
        Scene scene = Scene_GetStopScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        /* release timer */
        SelectA_ReleaseStopTimer(me->stopTimer);
        me->stopTimer = NULL;
        return TRUE;
    }
    return FALSE;
}

/*..........................................................................*/
void Group_ctor(Group *me)
{
    QHsm_ctor(&me->super, (QStateHandler)&Group_initial);

    GroupOffStates[0] = GetStateHandlerOffsetFromState(&Group_room_sensitive);
    GroupOffStates[1] = GetStateHandlerOffsetFromState(&Group_off);
}

/* HSM definition */

/*..........................................................................*/
static QState Group_initial(Group *me, QEvent const *e)
{
    /* QS dictionary entries */
    QS_FUN_DICTIONARY(&Group_room_sensitive);

    QS_FUN_DICTIONARY(&Group_off);
    QS_FUN_DICTIONARY(&Light_scene_0);
    QS_FUN_DICTIONARY(&Group_scene_0);

    QS_FUN_DICTIONARY(&Group_on);
    QS_FUN_DICTIONARY(&Light_scene_1);
    QS_FUN_DICTIONARY(&Group_scene_1);

    QS_FUN_DICTIONARY(&Group_dimming_down);
    QS_FUN_DICTIONARY(&Group_dimming_up);

    return Q_TRAN(&Group_room_sensitive);
}

/*..........................................................................*/
static QState Group_room_sensitive(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_UP_TIP_1X_SIG:
        /* fall through */
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Group_SendSTOP(me, e)) {
            return Q_HANDLED();
        }
        if (me->setLocalPriority) {
            return Q_TRAN(&Group_on_local_prio);
        } else {
            return Q_TRAN(&Group_on_local);
        }
    }
    case CT_2WAY_DOWN_TIP_1X_SIG: {
        if (Group_SendSTOP(me, e)) {
            return Q_HANDLED();
        }
        return Q_TRAN(&Group_off);
    }
    case CT_2WAY_UP_HOLD_START_SIG:
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        me->callMinDim = TRUE;
        return Q_TRAN(&Group_room_on_dimming_up);
    }
    case CT_2WAY_DOWN_HOLD_START_SIG:
        if (me->groupId == SHADE_GROUP) {
            me->callMinDim = TRUE;
            return Q_TRAN(&Group_room_on_dimming_down);
        }
    case CT_HOLD_START_SIG: {
        /* post event for main HSM */
        SceneBasicEvt * sceneBasicEvt = Q_NEW(SceneBasicEvt, BUTTON_GROUP_LOCAL_PRIO_OFF_SIG);
        sceneBasicEvt->sourceGroupId = me->groupId;
        sceneBasicEvt->sourceZoneId = me->localZoneId;
        sceneBasicEvt->sourceDeviceId = me->deviceId;
        sceneBasicEvt->outputChannel = me->outputChannel;
        QActive_postLIFO(me->parent, (QEvent *)sceneBasicEvt);
        return Q_HANDLED();
    }
    case CT_2WAY_DOWN_TIP_2X_SIG:
    case CT_2WAY_DOWN_TIP_3X_SIG:
    case CT_2WAY_DOWN_CLICK_3X_SIG:
    case CT_2WAY_DOWN_TIP_4X_SIG: {
        Group_StopSTOPTimer(me);
        /* post event for main HSM */
        QSignal sig;
        if (e->sig == CT_2WAY_DOWN_TIP_2X_SIG) {
            sig = BUTTON_GROUP_FORCE_SCENE_0_SIG;
        } else if (e->sig == CT_2WAY_DOWN_TIP_3X_SIG) {
            sig = BUTTON_GROUP_STANDBY_SIG;
        } else if (e->sig == CT_2WAY_DOWN_CLICK_3X_SIG) {
            sig = BUTTON_GROUP_FORCE_SCENE_0_AND_STANDBY_SIG;
        } else {
            sig = BUTTON_GROUP_DEEP_OFF_SIG;
        }
        SceneBasicEvt * sceneBasicEvt = Q_NEW(SceneBasicEvt, sig);
        sceneBasicEvt->sourceGroupId = me->groupId;
        sceneBasicEvt->sourceZoneId = me->localZoneId;
        sceneBasicEvt->sourceDeviceId = me->deviceId;
        sceneBasicEvt->outputChannel = me->outputChannel;
        QActive_postLIFO(me->parent, (QEvent *)sceneBasicEvt);
        return Q_HANDLED();
    }
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
        if (me->groupId == SHADE_GROUP) {
            me->callMinDim = TRUE;
            return Q_TRAN(&Group_room_on_dimming_down);
        }
    case CT_2WAY_DOWN_HOLD_END_SIG:
    case CT_2WAY_UP_TIP_2X_SIG:
    case CT_2WAY_UP_TIP_3X_SIG:
    case CT_2WAY_UP_TIP_4X_SIG:
    case CT_CLICK_2X_SIG:
    case CT_TIP_2X_SIG:
    case CT_CLICK_3X_SIG:
    case CT_TIP_3X_SIG:
    case CT_TIP_4X_SIG:
    case CT_HOLD_END_SIG:
    case CT_HOLD_REPEAT_SIG: {
        Group_StopSTOPTimer(me);
        {
            /* re-post event for main HSM */
            ClickForwardEvt * newEvent = Q_NEW(ClickForwardEvt, INTERNAL_CLICK_FORWARD_SIG);
            newEvent->groupId = me->groupId;
            newEvent->localZoneId = me->localZoneId;
            newEvent->event = e->sig;
            newEvent->buttonId = (me->buttonId >= BUTTON_GROUP_5) ? (me->buttonId - 4) : 0;
            newEvent->fid = me->fid;
            newEvent->key = me->key;
            newEvent->deviceId = me->deviceId;
            newEvent->apartmentWide = FALSE;
            newEvent->outputChannel = me->outputChannel;
            QActive_postLIFO(me->parent, (QEvent *)newEvent);
        }
        return Q_HANDLED();
    }
    case CT_SHORT_LONG_SIG:
    case CT_LOCAL_SHORT_LONG_SIG: {
        SelectA_SendGeneralAck();
        return Q_HANDLED();
    }
    case CALL_SCENE_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        bool force = sceneZoneGroupEvt->basic.force;
        SelectA_SendSceneZoneGroup(me->localZoneId, me->groupId, scene, force);
        /* fall through */
    }
    case CALL_SCENE_BROADCAST_SIG: {
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        Scene scene = (Scene)sceneBasicEvt->sceneId;
        if (Scene_IsOffScene(scene)) {
            Group_RestartSTOPTimer(me);
            return Q_TRAN(&Group_room_sensitive);
        } else if (Scene_IsStopScene(scene)) {
            Group_StopSTOPTimer(me);
        } else if (scene == Scene_GetOnScene(me->buttonId)) {
            Group_RestartSTOPTimer(me);
            return Q_TRAN(&Group_on);
        } else {
            Group_RestartSTOPTimer(me);
            return Q_TRAN(&Group_room_on);
        }
        /* don't change state */
        return Q_HANDLED();
    }
    case CALL_ROOM_ON_SIG: {
        return Q_TRAN(&Group_room_on);
    }
    case CALL_ROOM_OFF_SIG: {
        Group_StopSTOPTimer(me);
        return Q_TRAN(&Group_room_sensitive);
    }
    case CALL_ROOM_STOP_SIG: {
        Group_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Group_off(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Scene scene = Scene_GetOffScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        Group_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_TIP_1X_SIG:
        /* fall through */
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Group_SendSTOP(me, e)) {
            return Q_HANDLED();
        }
        break;
    }
    }
    return Q_SUPER(&Group_room_sensitive);
}

/*..........................................................................*/
static QState Group_room_on(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timer */
        Op_QTimeEvt_disarm(me->opTimer1);
        return Q_HANDLED();
    }
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Group_SendSTOP(me, e)) {
            return Q_HANDLED();
        }
        /* off scene */
        return Q_TRAN(&Group_off);
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        if (SelectA_RoomTimerRunning(me->localZoneId, me->groupId)) {
            SelectA_StopRoomTimer(me->localZoneId, me->groupId);
            Scene scene = Scene_GetStopScene(BUTTON_ZONE_0);
            SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        }
        return Q_TRAN(&Group_room_on_dimming_down);
    }
    case CT_2WAY_UP_HOLD_START_SIG:
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        return Q_TRAN(&Group_room_on_dimming_up);
    }
    case CT_2WAY_DOWN_HOLD_START_SIG:
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        return Q_TRAN(&Group_room_on_dimming_down);
    }
    case SET_SCENE_LOCAL_PRIO_BROADCAST_SIG: {
        return Q_TRAN(&Group_dim_down);
    }
    case CT_SHORT_LONG_SIG: {
        /* re-post event for main HSM */
        ClickForwardEvt * newEvent = Q_NEW(ClickForwardEvt, INTERNAL_CLICK_FORWARD_SIG);
        newEvent->groupId = me->groupId;
        newEvent->localZoneId = me->localZoneId;
        newEvent->event = e->sig;
        newEvent->buttonId = (me->buttonId >= BUTTON_GROUP_5) ? (me->buttonId - 4) : 0;
        newEvent->fid = me->fid;
        newEvent->key = me->key;
        newEvent->deviceId = me->deviceId;
        newEvent->apartmentWide = FALSE;
        newEvent->outputChannel = me->outputChannel;
        QActive_postLIFO(me->parent, (QEvent *)newEvent);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Group_room_sensitive);
}

/*..........................................................................*/
static QState Group_room_dimming(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_EXIT_SIG: {
        /* make sure call min-dim flag is reseted when leaving dim states */
        me->callMinDim = FALSE;
        return Q_HANDLED();
    }
    case CT_2WAY_UP_HOLD_START_SIG:
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        return Q_TRAN(&Group_room_on_dimming_up);
    }
    case CT_2WAY_DOWN_HOLD_START_SIG:
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        return Q_TRAN(&Group_room_on_dimming_down);
    }
    }
    return Q_SUPER(&Group_room_on);
}

/*..........................................................................*/
static QState Group_room_on_dim_down(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_EXIT_SIG: {
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        return Q_TRAN(&Group_room_on_dimming_down);
    }
    case CT_2WAY_DOWN_HOLD_END_SIG:
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_DEC_SIG: {
        return Q_TRAN(&Group_room_on_dim_up);
    }
    }
    return Q_SUPER(&Group_room_dimming);
}

/*..........................................................................*/
static QState Group_room_on_dimming_down(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* call scene DEC and arm timers */
        Scene scene = Scene_GetDecScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_DIM_TIMEOUT_TICKS, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Group_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_DEC_SIG: {
        /* call scene DEC (continue) and re-arm timer */
        Scene scene = Scene_GetDecContinueScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_DIM_TIMEOUT_TICKS, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG: {
        return Q_TRAN(&Group_room_on_dimming_up);
    }
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Group_room_on_dim_down);
}

/*..........................................................................*/
static QState Group_room_on_dim_up(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_EXIT_SIG: {
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        return Q_TRAN(&Group_room_on_dimming_up);
    }
    case CT_2WAY_UP_HOLD_END_SIG:
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_INC_SIG: {
        return Q_TRAN(&Group_room_on_dim_down);
    }
    }
    return Q_SUPER(&Group_room_dimming);
}

/*..........................................................................*/
static QState Group_room_on_dimming_up(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvtCtr dimTimeout = 2 * LIGHT_SHORT_MESSAGE_TIMEOUT_TICKS;
        if (me->callMinDim) {
            uint8_t numShortDSMsgs;
            /* call CallSceneMin_A with scene as parameter */
            Scene scene = Scene_GetOnScene(me->buttonId);
            numShortDSMsgs = SelectA_CallSceneMinZoneGroup(me->localZoneId, me->groupId, scene, me->deviceId);
            me->callMinDim = FALSE;
            dimTimeout += (numShortDSMsgs * LIGHT_SHORT_MESSAGE_TIMEOUT_TICKS);
        }
        /* call scene INC and arm timers */
        Scene scene = Scene_GetIncScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, dimTimeout, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Group_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timer2 */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_INC_SIG: {
        /* call scene INC (continue) and re-arm timer */
        Scene scene = Scene_GetIncContinueScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_DIM_TIMEOUT_TICKS, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG: {
        return Q_TRAN(&Group_room_on_dimming_up);
    }
    case CT_2WAY_UP_HOLD_REPEAT_SIG:
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Group_room_on_dim_up);
}

/*..........................................................................*/
static QState Group_on(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Group_RestartSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_TIP_1X_SIG:
        /* fall through */
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Group_SendSTOP(me, e)) {
            return Q_HANDLED();
        }
        break;
    }
    case CT_SHORT_LONG_SIG: {
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = me->deviceId;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 1;
            me->progMode->setLocalPriority = FALSE;

            /* re-post event for main HSM */
            SceneBasicEvt * sceneBasicEvt = Q_NEW(SceneBasicEvt, BUTTON_GROUP_PROGRAMM_MODE_SIG);
            sceneBasicEvt->sourceGroupId = me->groupId;
            sceneBasicEvt->sourceZoneId = me->localZoneId;
            sceneBasicEvt->sourceDeviceId = me->deviceId;
            sceneBasicEvt->sceneId = Scene_GetOnScene(me->buttonId);
            sceneBasicEvt->outputChannel = me->outputChannel;
            QActive_postLIFO(me->parent, (QEvent *)sceneBasicEvt);
        }
        return Q_HANDLED();
    }
    case REGISTER_SCENE_LOCAL_PRIO_BROADCAST_SIG: {
        return Q_TRAN(&Group_on_register_local_prio);
    }
    }
    return Q_SUPER(&Group_room_on);
}

/*..........................................................................*/
static QState Group_on_register_local_prio(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* register local priority */
        SelectA_RegisterLocalPrio(me->localZoneId, me->groupId, me->buttonId, 0);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* unregister set local priority */
        SelectA_UnregisterLocalPrio(me->localZoneId, me->groupId, me->buttonId);
        return Q_HANDLED();
    }
    case CALL_ROOM_SCENE_SIG: {
        /* send set local priority on scene 1 */
        Scene scene = Scene_GetOnScene(me->buttonId);
        SelectA_SetSceneLocalPrioZoneGroup(me->localZoneId, me->groupId, scene, me->deviceId);
        return Q_TRAN(&Group_dim_down);
    }
    }
    return Q_SUPER(&Group_on);
}

/*..........................................................................*/
static QState Group_on_local(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        Scene scene = Scene_GetOnScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Group_on);
}

/*..........................................................................*/
static QState Group_on_local_prio(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send register local priority to remote dSMs */
        DsmApi_SendZoneGroupActionRegisterSceneLocalPrioRequest(me->localZoneId, me->groupId, Scene_GetOnScene(me->buttonId), me->deviceId);
        /* register local priority */
        SelectA_RegisterLocalPrio(me->localZoneId, me->groupId, me->buttonId, me->deviceId);
        /* start timer */
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_SET_LOCAL_PRIO_TIMEOUT_TICKS, SET_SCENE_LOCAL_PRIO_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        Op_QTimeEvt_disarm(me->opTimer1);
        /* unregister local priority */
        SelectA_UnregisterLocalPrio(me->localZoneId, me->groupId, me->buttonId);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_TIP_2X_SIG:
    case CT_2WAY_UP_TIP_3X_SIG:
    case CT_2WAY_UP_TIP_4X_SIG: {
        /* send set local priority on scene 1 */
        Scene scene = Scene_GetOffScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        break;
    }
    case CT_2WAY_UP_HOLD_START_SIG:
    case CT_2WAY_UP_HOLD_REPEAT_SIG:
    case CT_2WAY_DOWN_HOLD_START_SIG:
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        /* send set local priority on scene 1 */
        Scene scene = Scene_GetOnScene(me->buttonId);
        SelectA_SetSceneLocalPrioZoneGroup(me->localZoneId, me->groupId, scene, me->deviceId);
        break;
    }
    case CALL_ROOM_SCENE_SIG: {
        Op_QTimeEvt_disarm(me->opTimer1);
        /* fall through */
    }
    case SET_SCENE_LOCAL_PRIO_SIG: {
        /* send set local priority on scene 1 */
        Scene scene = Scene_GetOnScene(me->buttonId);
        SelectA_SetSceneLocalPrioZoneGroup(me->localZoneId, me->groupId, scene, me->deviceId);
        return Q_TRAN(&Group_dim_down);
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        /* send set local priority on scene 1 */
        Scene scene = Scene_GetOnScene(me->buttonId);
        SelectA_SetSceneLocalPrioZoneGroup(me->localZoneId, me->groupId, scene, me->deviceId);
        return Q_TRAN(&Group_dimming_down);
    }
    }
    return Q_SUPER(&Group_on_local);
}

/*..........................................................................*/
static QState Group_local_prio(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_DOWN_TIP_1X_SIG:
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        /* off scene */
        if (Group_SendSTOP(me, e)) {
            return Q_HANDLED();
        }
        return Q_TRAN(&Group_off);
    }
    case CT_2WAY_UP_TIP_1X_SIG: {
        if (Group_SendSTOP(me, e)) {
            return Q_HANDLED();
        }
        if (me->setLocalPriority) {
            return Q_TRAN(&Group_on_local_prio);
        } else {
            return Q_TRAN(&Group_on_local);
        }
    }
    case CT_2WAY_UP_HOLD_START_SIG:
    case CT_2WAY_UP_HOLD_REPEAT_SIG: {
        if (SelectA_RoomTimerRunning(me->localZoneId, me->groupId)) {
            SelectA_StopRoomTimer(me->localZoneId, me->groupId);
            Scene scene = Scene_GetStopScene(BUTTON_ZONE_0);
            SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        }
        return Q_TRAN(&Group_dimming_up);
    }
    case CT_2WAY_DOWN_HOLD_START_SIG:
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG: {
        if (SelectA_RoomTimerRunning(me->localZoneId, me->groupId)) {
            SelectA_StopRoomTimer(me->localZoneId, me->groupId);
            Scene scene = Scene_GetStopScene(BUTTON_ZONE_0);
            SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        }
        return Q_TRAN(&Group_dimming_down);
    }
    case CT_2WAY_DOWN_TIP_2X_SIG:
    case CT_2WAY_DOWN_TIP_3X_SIG:
    case CT_2WAY_DOWN_CLICK_3X_SIG:
    case CT_2WAY_DOWN_TIP_4X_SIG: {
        Group_StopSTOPTimer(me);
        /* post event for main HSM */
        QSignal sig;
        if (e->sig == CT_2WAY_DOWN_TIP_2X_SIG) {
            sig = BUTTON_GROUP_FORCE_SCENE_0_SIG;
        } else if (e->sig == CT_2WAY_DOWN_TIP_3X_SIG) {
            sig = BUTTON_GROUP_STANDBY_SIG;
        } else if (e->sig == CT_2WAY_DOWN_CLICK_3X_SIG) {
            sig = BUTTON_GROUP_FORCE_SCENE_0_AND_STANDBY_SIG;
        } else {
            sig = BUTTON_GROUP_DEEP_OFF_SIG;
        }
        SceneBasicEvt * sceneBasicEvt = Q_NEW(SceneBasicEvt, sig);
        sceneBasicEvt->sourceGroupId = me->groupId;
        sceneBasicEvt->sourceZoneId = me->localZoneId;
        sceneBasicEvt->sourceDeviceId = me->deviceId;
        sceneBasicEvt->outputChannel = me->outputChannel;
        QActive_postLIFO(me->parent, (QEvent *)sceneBasicEvt);
        return Q_HANDLED();
    }
    case CT_2WAY_UP_TIP_2X_SIG:
    case CT_2WAY_UP_TIP_3X_SIG:
    case CT_2WAY_UP_TIP_4X_SIG:
    case CT_CLICK_2X_SIG:
    case CT_TIP_2X_SIG:
    case CT_CLICK_3X_SIG:
    case CT_TIP_3X_SIG:
    case CT_TIP_4X_SIG: {
        {
            Scene scene = Scene_GetOffScene(me->buttonId);
            SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
            Group_StopSTOPTimer(me);
        }
        /* re-post event for main HSM */
        ClickForwardEvt * newEvent = Q_NEW(ClickForwardEvt, INTERNAL_CLICK_FORWARD_SIG);
        newEvent->groupId = me->groupId;
        newEvent->localZoneId = me->localZoneId;
        newEvent->event = e->sig;
        newEvent->buttonId = (me->buttonId >= BUTTON_GROUP_5) ? (me->buttonId - 4) : 0;
        newEvent->fid = me->fid;
        newEvent->key = me->key;
        newEvent->deviceId = me->deviceId;
        newEvent->apartmentWide = FALSE;
        newEvent->outputChannel = me->outputChannel;
        QActive_postLIFO(me->parent, (QEvent *)newEvent);
        return Q_TRAN(&Group_room_on);
    }
    case CALL_SCENE_ZONE_GROUP_SIG: {
        SceneZoneGroupEvt * const sceneZoneGroupEvt = (SceneZoneGroupEvt * const)e;
        Scene scene = (Scene)sceneZoneGroupEvt->basic.sceneId;
        bool force = sceneZoneGroupEvt->basic.force;
        SelectA_SendSceneZoneGroup(me->localZoneId, me->groupId, scene, force);
        /* fall through */
    }
    case CALL_SCENE_BROADCAST_SIG: {
        SceneBasicEvt const * sceneBasicEvt = (SceneBasicEvt const *)e;
        bool force = sceneBasicEvt->force;
        Scene scene = (Scene)sceneBasicEvt->sceneId;
        if (Scene_IsOffScene(scene) || force) {
            Group_RestartSTOPTimer(me);
            return Q_TRAN(&Group_room_sensitive);
        } else if (Scene_IsStopScene(scene)) {
            Group_StopSTOPTimer(me);
        }
        /* don't change state */
        return Q_HANDLED();
    }
    case CALL_ROOM_OFF_SIG: {
        ClickForwardEvt const * clickForwardEvt = (ClickForwardEvt const *)e;
        if (clickForwardEvt->force) {
            Group_StopSTOPTimer(me);
            return Q_TRAN(&Group_room_sensitive);
        }
        /* don't change state */
        return Q_HANDLED();
    }
    case CALL_ROOM_ON_SIG: {
        ClickForwardEvt const * clickForwardEvt = (ClickForwardEvt const *)e;
        if (clickForwardEvt->force) {
            return Q_TRAN(&Group_room_on);
        }
        /* don't change state */
        return Q_HANDLED();
    }
    case CALL_ROOM_STOP_SIG: {
        Group_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case CT_SHORT_LONG_SIG: {
        if (me->progMode->localZoneId == 0) {
            /* save initiating device ID, button ID and scene */
            me->progMode->initiatingDeviceId = me->deviceId;
            me->progMode->buttonId = me->buttonId;
            me->progMode->sceneIndex = 1;
            me->progMode->setLocalPriority = TRUE;

            /* re-post event for main HSM */
            SceneBasicEvt * sceneBasicEvt = Q_NEW(SceneBasicEvt, BUTTON_GROUP_PROGRAMM_MODE_SIG);
            sceneBasicEvt->sourceGroupId = me->groupId;
            sceneBasicEvt->sourceZoneId = me->localZoneId;
            sceneBasicEvt->sourceDeviceId = me->deviceId;
            sceneBasicEvt->sceneId = Scene_GetOnScene(me->buttonId);
            sceneBasicEvt->outputChannel = me->outputChannel;
            QActive_postLIFO(me->parent, (QEvent *)sceneBasicEvt);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState Group_dim_down(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_2WAY_UP_TIP_1X_SIG:
        /* fall through */
    case CT_TIP_1X_SIG:
        /* fall through */
    case CT_CLICK_1X_SIG: {
        if (Group_SendSTOP(me, e)) {
            return Q_HANDLED();
        }
        break;
    }
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        if (SelectA_RoomTimerRunning(me->localZoneId, me->groupId)) {
            SelectA_StopRoomTimer(me->localZoneId, me->groupId);
            Scene scene = Scene_GetStopScene(BUTTON_ZONE_0);
            SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        }
        return Q_TRAN(&Group_dimming_down);
    }
    case CT_2WAY_DOWN_HOLD_END_SIG:
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_DEC_SIG: {
        return Q_TRAN(&Group_dim_up);
    }
    }
    return Q_SUPER(&Group_local_prio);
}

/*..........................................................................*/
static QState Group_dimming_down(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* call scene DEC and arm timers */
        Scene scene = Scene_GetDecScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_DIM_TIMEOUT_TICKS, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Group_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timers */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_DEC_SIG: {
        /* call scene DEC (continue) and re-arm timer */
        Scene scene = Scene_GetDecContinueScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_DIM_TIMEOUT_TICKS, CALL_DEC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG: {
        return Q_TRAN(&Group_dimming_up);
    }
    case CT_2WAY_DOWN_HOLD_REPEAT_SIG:
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Group_dim_down);
}

/*..........................................................................*/
static QState Group_dim_up(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case CT_HOLD_START_SIG:
        /* fall through */
    case CT_HOLD_REPEAT_SIG: {
        return Q_TRAN(&Group_dimming_up);
    }
    case CT_2WAY_UP_HOLD_END_SIG:
    case CT_HOLD_END_SIG: {
        /* send general acknowledge */
        SelectA_SendGeneralAck();
        /* fall through */
    }
    case END_INC_SIG: {
        return Q_TRAN(&Group_dim_down);
    }
    }
    return Q_SUPER(&Group_local_prio);
}

/*..........................................................................*/
static QState Group_dimming_up(Group *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* call scene INC and arm timers */
        Scene scene = Scene_GetIncScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_DIM_TIMEOUT_TICKS, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Op_QTimeEvt_postIn(me->opTimer2, me->parent, LIGHT_HOLD_END_TIMEOUT_TICKS, END_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        Group_StopSTOPTimer(me);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* disarm timer2 */
        Op_QTimeEvt_disarm(me->opTimer1);
        Op_QTimeEvt_disarm(me->opTimer2);
        return Q_HANDLED();
    }
    case CALL_INC_SIG: {
        /* call scene INC (continue) and re-arm timer */
        Scene scene = Scene_GetIncContinueScene(me->buttonId);
        SelectA_CallSceneZoneGroup(me->localZoneId, me->groupId, me->localZoneId, me->groupId, scene, me->deviceId);
        Op_QTimeEvt_postIn(me->opTimer1, me->parent, LIGHT_DIM_TIMEOUT_TICKS, CALL_INC_SIG, me->buttonId, me->localZoneId, me->groupId, me->deviceId, FALSE, me->outputChannel);
        return Q_HANDLED();
    }
    case CT_HOLD_START_SIG: {
        return Q_TRAN(&Group_dimming_up);
    }
    case CT_2WAY_UP_HOLD_REPEAT_SIG:
    case CT_HOLD_REPEAT_SIG: {
        /* re-arm timer */
        Op_QTimeEvt_rearm(me->opTimer2, LIGHT_HOLD_END_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&Group_dim_up);
}
