/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.
    

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <stdlib.h>

#include "util.h"
#include "msgutil.h"

bool msgutil_get_int(const Vdcapi__PropertyValue *msg, int64_t *out)
{
    if (!msg) {
        return FALSE;
    }

    if (!msg->has_v_int64) {
        return FALSE;
    }

    if (out) {
        *out = msg->v_int64;
    }

    return TRUE;
}

bool msgutil_get_uint(const Vdcapi__PropertyValue *msg, uint64_t *out)
{
    if (!msg) {
        return FALSE;
    }

    if (!msg->has_v_uint64) {
        return FALSE;
    }

    if (out) {
        *out = msg->v_uint64;
    }

    return TRUE;
}

bool msgutil_get_bool(const Vdcapi__PropertyValue *msg, bool *out)
{
    if (!msg) {
        return FALSE;
    }

    if (!msg->has_v_bool) {
        return FALSE;
    }

    if (out) {
        *out = msg->v_bool;
    }

    return TRUE;
}

bool msgutil_get_double(const Vdcapi__PropertyValue *msg, double *out)
{
    if (!msg) {
        return FALSE;
    }

    if (!msg->has_v_double) {
        return FALSE;
    }

    if (out) {
        *out = msg->v_double;
    }

    return TRUE;
}

bool msgutil_get_string(const Vdcapi__PropertyValue *msg, const char **out)
{
    if (!msg) {
        return FALSE;
    }

    if (!msg->v_string) {
        return FALSE;
    }

    if (out) {
        *out = msg->v_string;
    }

    return TRUE;
}

Vdcapi__PropertyValue *msgutil_get_property(
        Vdcapi__VdcResponseGetProperty *msg, const char *name)
{
    size_t i;
    Vdcapi__PropertyValue *val = NULL;

    if (msg && (msg->n_properties > 0)) {
        for (i = 0; i < msg->n_properties; ++i) {
            if (strcmp(msg->properties[i]->name, name) == 0) {
                val = msg->properties[i]->value;
                break;
            }
        } // for
    }

    return val;
}

Vdcapi__PropertyValue *msgutil_get_push_property(
        Vdcapi__VdcSendPushProperty *msg, const char *name)
{
    size_t i;
    Vdcapi__PropertyValue *val = NULL;

    if (msg && (msg->n_properties > 0)) {
        for (i = 0; i < msg->n_properties; ++i) {
            if (strcmp(msg->properties[i]->name, name) == 0) {
                val = msg->properties[i]->value;
                break;
            }
        } // for
    }

    return val;
}

bool msgutil_get_int_property(Vdcapi__VdcResponseGetProperty *prop,
                              const char *name,
                              int64_t min, int64_t max, int64_t *out)
{
    Vdcapi__PropertyValue *val = msgutil_get_property(prop, name);
    if (!val) {
        vdsm_report(LOG_ERR, "error getting property %s: missing "
               "value\n", name);
        return FALSE;
    }

    if (!msgutil_get_int(val, out)) {
        vdsm_report(LOG_ERR, "error getting property %s: wrong "
               "value type\n", name);
        return FALSE;
    }

    if ((*out > max) || (*out < min)) {
        vdsm_report(LOG_ERR, "error getting property %s: value out of "
               "bounds\n", name);
        return FALSE;
    }

    return TRUE;
}

bool msgutil_get_uint_property(Vdcapi__VdcResponseGetProperty *prop,
                               const char *name,
                               uint64_t max, uint64_t *out)
{
    Vdcapi__PropertyValue *val = msgutil_get_property(prop, name);
    if (!val) {
        vdsm_report(LOG_ERR, "error getting property %s: missing "
               "value\n", name);
        return FALSE;
    }

    if (!msgutil_get_uint(val, out)) {
        vdsm_report(LOG_ERR, "error getting property %s: wrong "
               "value type\n", name);
        return FALSE;
    }

    if (*out > max) {
        vdsm_report(LOG_ERR, "error getting property %s: value out of "
               "bounds\n", name);
        return FALSE;
    }

    return TRUE;
}

bool msgutil_get_double_property(Vdcapi__VdcResponseGetProperty *prop,
                                 const char *name, double *out)
{
    Vdcapi__PropertyValue *val = msgutil_get_property(prop, name);
    if (!val) {
        vdsm_report(LOG_ERR, "error getting property %s: missing "
               "value\n", name);
        return FALSE;
    }

    if (!msgutil_get_double(val, out)) {
        vdsm_report(LOG_ERR, "error getting property %s: wrong "
               "value type\n", name);
        return FALSE;
    }

    return TRUE;
}

bool msgutil_get_bool_property(Vdcapi__VdcResponseGetProperty *prop,
                               const char *name, bool *out)
{
    Vdcapi__PropertyValue *val = msgutil_get_property(prop, name);
    if (!val) {
        vdsm_report(LOG_ERR, "error getting property %s: missing "
               "value\n", name);
        return FALSE;
    }

    if (!msgutil_get_bool(val, out)) {
        vdsm_report(LOG_ERR, "error getting property %s: wrong "
               "value type\n", name);
        return FALSE;
    }

    return TRUE;
}

bool msgutil_get_string_property(Vdcapi__VdcResponseGetProperty *prop,
                                 const char *name, const char **out)
{
    Vdcapi__PropertyValue *val = msgutil_get_property(prop, name);
    if (!val) {
        vdsm_report(LOG_ERR, "error getting property %s: missing "
               "value\n", name);
        return FALSE;
    }

    if (!msgutil_get_string(val, out)) {
        vdsm_report(LOG_ERR, "error getting property %s: wrong "
               "value type\n", name);
        return FALSE;
    }

    return TRUE;
}

